/* drivers/barcelona/bat/bat.c
 *
 * Implementation of the battery driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */ 
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/hardware/clock.h>
#include <barcelona/Barc_Battery.h>
#include <barcelona/Barc_adc.h>
#include <barcelona/debug.h>
#include <barcelona/gopins.h>
#include <barcelona/usbmode.h>

/* defines */
#define PFX "bat: "
#define PK_DBG PK_DBG_FUNC

static int bat_disabled;

#define BATTERY_AVERAGE_SIZE 50

static unsigned short batteryVoltageBuffer[BATTERY_AVERAGE_SIZE];
static unsigned short batteryCurrentBuffer[BATTERY_AVERAGE_SIZE];
static int batteryIndex = 0;
static int batteryWait = BATTERY_AVERAGE_SIZE;
static int batteryVoltageSum = 0;
static int batteryCurrentSum = 0;
static int voltageSource = VOLTAGE_SOURCE_BATTERY;

/**
 * Based on the usb mode change the power bypass
 **/
void  usb_state_change_listener(USB_STATE previous_state,USB_STATE current_state, void* arg){
        if (IO_HaveUsbBusPowered()){
                switch (previous_state){
                case USB_STATE_INITIAL:
                case USB_STATE_IDLE:
                case USB_STATE_DEVICE_DETECT:
                case USB_STATE_DEVICE:
                        /* if we where in a now powerd mode and are switching to a powered mode...*/
                        if (current_state == USB_STATE_HOST_DETECT || current_state ==  USB_STATE_HOST || current_state == USB_STATE_CLA){
                                printk(KERN_INFO PFX " %s ENABLE POWER CHARGE \n", __FUNCTION__);
                                IO_Activate(USB_PWR_BYPASS);
                        }
                        break;
                case USB_STATE_HOST_DETECT:
                case USB_STATE_HOST:
                case USB_STATE_CLA:
                        if (current_state == USB_STATE_INITIAL || current_state ==  USB_STATE_IDLE || current_state == USB_STATE_DEVICE_DETECT || current_state == USB_STATE_DEVICE ){
                                printk(KERN_INFO PFX " %s DISABLE POWER CHARGE \n", __FUNCTION__);
                                IO_Deactivate(USB_PWR_BYPASS);
                        }

                        break;
                }
        }

}

static void bat_hw_init(void) 
{
	int i;

	IO_SetInput(CHARGEFAULT);
	IO_SetInput(CHARGING);
	IO_SetInput(ACPWR);
	IO_Deactivate(CHARGE_OUT);

	for (i=0; i<BATTERY_AVERAGE_SIZE; i++) {
		batteryCurrentBuffer[i] = 0;
		batteryVoltageBuffer[i] = 0;
	}
	batteryVoltageSum = 0;
	batteryCurrentSum = 0;
	batteryWait = BATTERY_AVERAGE_SIZE;
}

static void bat_adc_poll(short adcBuffer[ADC_CHANNELS], void* arg )
{
	unsigned int alc = 0;
	unsigned int bat = (adcBuffer[ADC_BAT] * 2450) / adcBuffer[ADC_REF];
	unsigned int sample;

	// voltage = (1.2 * strBatteryStatus.u16BatteryVoltage * 2 / strBatteryStatus.u16ReferenceVoltage);
	// ampere = (1200.0 * strBatteryStatus.u16ChargeCurrent / (strBatteryStatus.u16ReferenceVoltage * 1500));
	
	if( IO_HaveAlkaline() ){
		alc = (adcBuffer[ADC_ALCALINE] * 2450) / adcBuffer[ADC_REF];
	}

	if (alc > 1500) {
		sample = alc;
		voltageSource = VOLTAGE_SOURCE_BATTERY_PACK;
	} else {
		sample = bat;
		voltageSource = VOLTAGE_SOURCE_BATTERY;
	}

	batteryVoltageSum -= batteryVoltageBuffer[batteryIndex];
	batteryVoltageSum += sample;
	batteryVoltageBuffer[batteryIndex] = sample;

	if (voltageSource == VOLTAGE_SOURCE_BATTERY) {
		sample = (adcBuffer[ADC_CHRG] * 942) / adcBuffer[ADC_REF];
		batteryCurrentSum -= batteryCurrentBuffer[batteryIndex];
		batteryCurrentSum += sample;
		batteryCurrentBuffer[batteryIndex] = sample;
	}

	batteryIndex++;
	if (batteryIndex >= BATTERY_AVERAGE_SIZE) batteryIndex = 0;
	if (batteryWait > 0) batteryWait--;
}

static int bat_get_voltage(void)
{
	return batteryVoltageSum / BATTERY_AVERAGE_SIZE;
}

static int bat_get_current(void)
{
	return batteryCurrentSum / BATTERY_AVERAGE_SIZE;
}

static int bat_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret;

	if (bat_disabled)
		return -ENODEV;

	switch (cmd) {
	case IO_ENABLE_CHARGING:
		IO_SetInput(BATT_TEMP_OVER);
		ret = 0;
		break;
	case IO_DISABLE_CHARGING:
		IO_Activate(BATT_TEMP_OVER);
		ret = 0;
		break;
	case IOR_BATTERY_STATUS:
		if (!access_ok(VERIFY_WRITE, (void __user *) arg, sizeof(BATTERY_STATUS))) {
			ret = -EFAULT;
		} else {
			BATTERY_STATUS bs;
			bs.u16BatteryVoltage   = bat_get_voltage();
			bs.u16ChargeCurrent    = bat_get_current();
			bs.u8VoltageSource     = voltageSource;

			// Power state
			switch (IO_GetChargerType()) {
				case GOCHARGER_LTC3455:
					// Use ADC to detect charge current
					if (IO_GetInput(ACPWR))
					{
						short buf[ADC_CHANNELS];
						adc_get_buffer(buf);
						bs.u8ChargeStatus = CHARGE_STATE_COMPLETE;
						if (buf[ADC_CHRG] > 50)  bs.u8ChargeStatus = CHARGE_STATE_TOPPINGUP;
						if (buf[ADC_CHRG] > 500) bs.u8ChargeStatus = CHARGE_STATE_FAST_CHARGING;
					} else {
						bs.u8ChargeStatus = CHARGE_STATE_ACPWR_OFF;
					}
					break;
				case GOCHARGER_LTC1733:
					// Use CHARGE_OUT to detect charging
					if (IO_GetInput(ACPWR)) {
						if (IO_GetInput(CHARGEFAULT)) {
							bs.u8ChargeStatus = CHARGE_STATE_FAULT;
						}
						else if (IO_GetInput(CHARGING)) {
							IO_Activate(CHARGE_OUT);
							if (IO_GetInput(CHARGING)) {
								bs.u8ChargeStatus = CHARGE_STATE_TOPPINGUP;
							} else {
								bs.u8ChargeStatus = CHARGE_STATE_FAST_CHARGING;
							}
							IO_Deactivate(CHARGE_OUT);
						} else {
							bs.u8ChargeStatus = CHARGE_STATE_COMPLETE;
						}
					} else {
						bs.u8ChargeStatus = CHARGE_STATE_ACPWR_OFF;
					}
					break;
				default:
					bs.u8ChargeStatus = CHARGE_STATE_ACPWR_OFF;
			}

			ret = __copy_to_user((void __user *) arg, &bs, sizeof(BATTERY_STATUS)) ? -EFAULT : 0;
		}
		break;
	default:
		PK_WARN("Invalid ioctl command %u\n", cmd);
		ret = -EINVAL;
		break;
	}

	return ret;
}

static int bat_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int bat_release(struct inode *inode, struct file *file)
{
	return 0;
}

/* Kernel interface */
static struct file_operations bat_fops = {
	.owner		= THIS_MODULE,
	.ioctl		= bat_ioctl,
	.open		= bat_open,
	.release	= bat_release,
};

static int bat_probe(struct device *dev)
{
	int ret;

	bat_hw_init();

	PK_DBG("Registering ADC pollfunc\n");
	ret = adc_register_poll(bat_adc_poll, NULL );
	if (ret < 0) {
		PK_ERR("Unable to register ADC pollfunc (%d)\n", ret);
		return ret;
	}

	PK_DBG("Registering chardev\n");
	ret = register_chrdev(BATTERY_MAJOR, BATTERY_DEVNAME, &bat_fops);
	if (ret != 0) {
		PK_ERR("Unable to register chardev on major=%d (%d)\n", BATTERY_MAJOR, ret);
		return ret;
	}

	PK_DBG("Registering usbmode callback\n");
        ret = add_usb_state_change_listener(usb_state_change_listener,0);
        if (ret != 0){
                printk(KERN_INFO PFX "%s ERROR failed to register usbmode callback\n", __FUNCTION__);
                return ret;
        }
	PK_DBG("Done\n");
	return 0;
}

static int bat_remove(struct device *dev)
{
	PK_DBG("Unregistering chardev\n");
	unregister_chrdev(BATTERY_MAJOR, BATTERY_DEVNAME);
	remove_usb_state_change_listener(usb_state_change_listener,0);

	PK_DBG("Done\n");
	return 0;
}

static void bat_shutdown(struct device *dev)
{
	PK_DBG("Shutting down\n");
	/* Nothing yet */
}

#ifdef CONFIG_PM

static int bat_suspend(struct device *dev, u32 state, u32 level)
{
	PK_DBG("dev = %p, state = %u, level = %u\n", dev, state, level);
	if (level == SUSPEND_POWER_DOWN) {
		bat_disabled = 1;
		if (IO_HaveUsbBusPowered()){
			IO_Deactivate(USB_PWR_BYPASS);
		}
	}

	return 0;
}

static int bat_resume(struct device *dev, u32 level)
{
	PK_DBG("dev = %p, level = %u\n", dev, level);
	if (level == RESUME_POWER_ON) {
		bat_hw_init();
		bat_disabled = 0;
		if (IO_HaveUsbBusPowered()){
			/*make sure we don't blow up our friend PC's */
			IO_Deactivate(USB_PWR_BYPASS);
		}
	}
	return 0;
}

#else /* CONFIG_PM */
#define bat_suspend NULL
#define bat_resume  NULL
#endif /* CONFIG_PM */

static struct device_driver bat_driver = {
	.name		= "tomtomgo-bat",
	.bus		= &platform_bus_type,
	.probe		= bat_probe,
	.remove		= bat_remove,
	.shutdown	= bat_shutdown,
	.suspend	= bat_suspend,
	.resume		= bat_resume,
};

static int __init bat_mod_init(void)
{
	int ret;

	printk(KERN_INFO "TomTom GO Battery Driver, (C) 2004,2005,2006,2007 TomTom BV\n");
	PK_DBG("Registering driver\n");
	ret = driver_register(&bat_driver);
	if (ret) {
		PK_ERR("Unable to register driver (%d)\n", ret);
		return ret;
	}
	PK_DBG("Done\n");
	return 0;
}

static void __exit bat_mod_exit(void)
{
	PK_DBG("Unregistering driver\n");
	driver_unregister(&bat_driver);
	PK_DBG("Done\n");
}

module_init(bat_mod_init);
module_exit(bat_mod_exit);

MODULE_AUTHOR("Dimitry Andric <dimitry.andric@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO Battery Driver");
MODULE_LICENSE("GPL");

/* EOF */
